import logging
import os
import pandas as pd
import pkg_resources


log = logging.getLogger(__name__)


legislation_directory = os.path.join(
    pkg_resources.get_distribution('openfisca_survey_manager').location,
    'openfisca_survey_manager',
    'assets',
    )


sub_levels = ['divisions', 'groupes', 'classes', 'sous_classes', 'postes']
divisions = ['0{}'.format(i) for i in range(1, 10)] + ['11', '12']


def build_coicop_level_nomenclature(level, keep_code = False, to_csv = False):
    assert level in sub_levels
    log.debug("Reading nomenclature coicop source data for level {}".format(level))
    try:
        data_frame = pd.read_csv(
            os.path.join(legislation_directory, 'nomenclature_coicop_source_by_{}.csv'.format(level)),
            sep = ';',
            header = None,
            )
    except Exception as e:
        log.info("Error when reading nomenclature coicop source data for level {}".format(level))
        raise(e)

    data_frame.reset_index(inplace = True)
    data_frame.rename(columns = {0: 'code_coicop', 1: 'label_{}'.format(level[:-1])}, inplace = True)
    data_frame = data_frame.iloc[2:].copy()

    index, stop = 0, False
    for sub_level in sub_levels:
        if stop:
            continue
        if sub_level == 'divisions':
            data_frame[sub_level] = data_frame['code_coicop'].str[index:index + 2].astype(int)
            index = index + 3
        else:
            data_frame[sub_level] = data_frame['code_coicop'].str[index:index + 1].astype(int)
            index = index + 2

        if level == sub_level:
            stop = True

    if keep_code or level == 'postes':
        data_frame['code_coicop'] = data_frame['code_coicop'].str.lstrip("0")
    else:
        del data_frame['code_coicop']

    data_frame.reset_index(inplace = True, drop = True)
    if to_csv:
        data_frame.to_csv(
            os.path.join(legislation_directory, 'nomenclature_coicop_by_{}.csv'.format(level)),
            )

    return data_frame


def build_raw_coicop_nomenclature():
    """Builds raw COICOP nomenclature"""
    for index in range(len(sub_levels) - 1):
        level = sub_levels[index]
        next_level = sub_levels[index + 1]
        on = sub_levels[:index + 1]
        if index == 0:
            coicop_nomenclature = pd.merge(
                build_coicop_level_nomenclature(level), build_coicop_level_nomenclature(next_level),
                on = on, left_index = False, right_index = False)
        else:
            coicop_nomenclature = pd.merge(coicop_nomenclature, build_coicop_level_nomenclature(next_level), on = on)

    coicop_nomenclature = coicop_nomenclature[
        ['code_coicop']
        + ['label_{}'.format(sub_level[:-1]) for sub_level in sub_levels]
        + sub_levels
        ].copy()

    return coicop_nomenclature[['label_division', 'label_groupe', 'label_classe',
       'label_sous_classe', 'label_poste', 'code_coicop']].copy()


if __name__ == "__main__":
    import sys
    logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    raw_coicop_nomenclature = build_raw_coicop_nomenclature()
    log.info(raw_coicop_nomenclature)





import os
import pkg_resources


from openfisca_survey_manager.survey_collections import SurveyCollection
from openfisca_survey_manager.scripts.build_collection import add_survey_to_collection

is_travis = 'TRAVIS' in os.environ
is_circleci = 'CIRCLECI' in os.environ


def test_add_survey_to_collection():
    if is_travis or is_circleci:
        return
    name = 'fake'
    survey_name = 'fake_survey'
    data_dir = os.path.join(
        pkg_resources.get_distribution('openfisca-survey-manager').location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        )
    survey_collection = SurveyCollection(name = name)
    saved_fake_survey_file_path = os.path.join(data_dir, 'help.sas7bdat')
    add_survey_to_collection(
        survey_name = survey_name,
        survey_collection = survey_collection,
        sas_files = [saved_fake_survey_file_path],
        stata_files = []
        )
    ordered_dict = survey_collection.to_json()
    assert survey_name in list(ordered_dict['surveys'].keys())




import shutil


import logging
import os
import pkg_resources


from openfisca_core.model_api import *  # noqa analysis:ignore
from openfisca_core import periods
from openfisca_core.tools import assert_near

from openfisca_country_template import CountryTaxBenefitSystem


from openfisca_survey_manager.input_dataframe_generator import (
    make_input_dataframe_by_entity,
    random_data_generator,
    randomly_init_variable,
    )
from openfisca_survey_manager.scenarios import AbstractSurveyScenario


log = logging.getLogger(__name__)


tax_benefit_system = CountryTaxBenefitSystem()


def create_randomly_initialized_survey_scenario(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000, collection = "test_random_generator", use_marginal_tax_rate = False):
    if collection is not None:
        return create_randomly_initialized_survey_scenario_from_table(
            nb_persons, nb_groups, salary_max_value, rent_max_value, collection, use_marginal_tax_rate)
    else:
        return create_randomly_initialized_survey_scenario_from_data_frame(
            nb_persons, nb_groups, salary_max_value, rent_max_value, use_marginal_tax_rate)


def create_randomly_initialized_survey_scenario_from_table(nb_persons, nb_groups, salary_max_value, rent_max_value, collection, use_marginal_tax_rate):
    variable_generators_by_period = {
        periods.period('2017-01'): [
            {
                'variable': 'salary',
                'max_value': salary_max_value,
                },
            {
                'variable': 'rent',
                'max_value': rent_max_value,
                }
            ],
        periods.period('2018-01'): [
            {
                'variable': 'salary',
                'max_value': salary_max_value,
                },
            ],
        }
    table_by_entity_by_period = random_data_generator(tax_benefit_system, nb_persons, nb_groups,
        variable_generators_by_period, collection)
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(tax_benefit_system = tax_benefit_system)
    survey_scenario.used_as_input_variables = ['salary', 'rent', 'housing_occupancy_status']
    survey_scenario.year = 2017
    survey_scenario.collection = collection
    data = {
        'survey': 'input',
        'input_data_table_by_entity_by_period': table_by_entity_by_period
        }
    survey_scenario.varying_variable = 'salary'
    survey_scenario.init_from_data(data = data, use_marginal_tax_rate = use_marginal_tax_rate)
    return survey_scenario


def create_randomly_initialized_survey_scenario_from_data_frame(nb_persons, nb_groups, salary_max_value, rent_max_value, use_marginal_tax_rate = False):
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(tax_benefit_system = tax_benefit_system)
    survey_scenario.year = 2017
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    period = periods.period('2017-01')
    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            }
        }
    survey_scenario.varying_variable = 'salary'
    survey_scenario.init_from_data(data = data, use_marginal_tax_rate = use_marginal_tax_rate)
    return survey_scenario


def generate_input_input_dataframe_by_entity(nb_persons, nb_groups, salary_max_value, rent_max_value):
    input_dataframe_by_entity = make_input_dataframe_by_entity(tax_benefit_system, nb_persons, nb_groups)
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        'salary',
        max_value = salary_max_value,
        condition = "household_role == 'first_parent'"
        )
    randomly_init_variable(
        tax_benefit_system,
        input_dataframe_by_entity,
        'rent',
        max_value = rent_max_value
        )
    return input_dataframe_by_entity


def test_input_dataframe_generator(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):
    input_dataframe_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    assert (input_dataframe_by_entity['person']['household_role'] == "first_parent").sum() == 5
    assert (input_dataframe_by_entity['person'].loc[
        input_dataframe_by_entity['person']['household_role'] != "first_parent",
        'salary'
        ] == 0).all()

    assert (input_dataframe_by_entity['person'].loc[
        input_dataframe_by_entity['person']['household_role'] == "first_parent",
        'salary'
        ] > 0).all()
    assert (input_dataframe_by_entity['person'].loc[
        input_dataframe_by_entity['person']['household_role'] == "first_parent",
        'salary'
        ] <= salary_max_value).all()

    assert (input_dataframe_by_entity['household']['rent'] > 0).all()
    assert (input_dataframe_by_entity['household']['rent'] < rent_max_value).all()


def test_survey_scenario_input_dataframe_import(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):

    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(tax_benefit_system = tax_benefit_system)
    survey_scenario.year = 2017
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    period = periods.period('2017-01')
    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            }
        }
    survey_scenario.init_from_data(data = data)

    simulation = survey_scenario.simulation
    assert (
        simulation.calculate('salary', period) == input_data_frame_by_entity['person']['salary']
        ).all()
    assert (
        simulation.calculate('rent', period) == input_data_frame_by_entity['household']['rent']
        ).all()


def test_survey_scenario_input_dataframe_import_scrambled_ids(nb_persons = 10, nb_groups = 5, salary_max_value = 50000,
        rent_max_value = 1000):
    input_data_frame_by_entity = generate_input_input_dataframe_by_entity(
        nb_persons, nb_groups, salary_max_value, rent_max_value)
    input_data_frame_by_entity['person']['household_id'] = 4 - input_data_frame_by_entity['person']['household_id']
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(tax_benefit_system = tax_benefit_system)
    survey_scenario.year = 2017
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    period = periods.period('2017-01')
    data = {
        'input_data_frame_by_entity_by_period': {
            period: input_data_frame_by_entity
            }
        }
    survey_scenario.init_from_data(data = data)
    simulation = survey_scenario.simulation
    period = periods.period('2017-01')
    assert (
        simulation.calculate('salary', period) == input_data_frame_by_entity['person']['salary']
        ).all()
    assert (
        simulation.calculate('rent', period) == input_data_frame_by_entity['household']['rent']
        ).all()


def test_dump_survey_scenario():
    survey_scenario = create_randomly_initialized_survey_scenario()
    directory = os.path.join(
        pkg_resources.get_distribution('openfisca-survey-manager').location,
        'openfisca_survey_manager',
        'tests',
        'data_files',
        'dump',
        )
    if os.path.exists(directory):
        shutil.rmtree(directory)
    survey_scenario.dump_simulations(directory = directory)
    df = survey_scenario.create_data_frame_by_entity(variables = ['salary', 'rent'])
    household = df['household']
    person = df['person']
    assert not household.empty
    assert not person.empty
    del survey_scenario
    survey_scenario = AbstractSurveyScenario()
    survey_scenario.set_tax_benefit_systems(tax_benefit_system = tax_benefit_system)
    survey_scenario.used_as_input_variables = ['salary', 'rent']
    survey_scenario.year = 2017
    survey_scenario.restore_simulations(directory = directory)
    df2 = survey_scenario.create_data_frame_by_entity(variables = ['salary', 'rent'], period = '2017-01')

    assert (df2['household'] == household).all().all()
    assert (df2['person'] == person).all().all()


def test_inflate():
    survey_scenario = create_randomly_initialized_survey_scenario()
    period = "2017-01"
    inflator = 2.42
    inflator_by_variable = {'rent': inflator}

    rent_before_inflate = survey_scenario.compute_aggregate('rent', period = period)
    survey_scenario.inflate(inflator_by_variable = inflator_by_variable, period = period)
    rent_after_inflate = survey_scenario.compute_aggregate('rent', period = period)

    assert_near(
        rent_after_inflate,
        inflator * rent_before_inflate,
        relative_error_margin = 1e-6,
        message = "Failing inflate with inflator_by_variable: rent_after_inflate = {} != {} = rent_before_inflate ({}) x inflator ({})".format(
            rent_after_inflate,
            rent_before_inflate * inflator,
            rent_before_inflate,
            inflator
            )
        )

    target = 3e5
    target_by_variable = {'salary': target}
    salary_before_inflate = survey_scenario.compute_aggregate('salary', period = period)
    survey_scenario.inflate(target_by_variable = target_by_variable, period = period)
    salary_after_inflate = survey_scenario.compute_aggregate('salary', period = period)
    assert_near(
        salary_after_inflate,
        target,
        relative_error_margin = 1e-6,
        message = "Failing inflate with inflator_by_variable: salary_after_inflate = {} != {} = target (salary_before_inflate = {})\n".format(
            salary_after_inflate,
            target,
            salary_before_inflate,
            )
        )


if __name__ == "__main__":
    import sys
    log = logging.getLogger(__name__)
    logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)
    test_inflate()
    # test_create_data_frame_by_entity()





from openfisca_core import periods
from openfisca_country_template import CountryTaxBenefitSystem
from openfisca_survey_manager.utils import inflate_parameters, parameters_asof


def test_asof_simple_annual_parameter():
    """
        Test parameters_asof on a simple parameter
    """
    tax_benefit_system = CountryTaxBenefitSystem()
    parameters = tax_benefit_system.parameters
    income_tax_rate_2014 = parameters.taxes.income_tax_rate(2014)
    income_tax_rate_2015 = parameters.taxes.income_tax_rate(2015)
    assert income_tax_rate_2015 != income_tax_rate_2014
    parameters_asof(parameters, instant = "2014")
    assert parameters.taxes.income_tax_rate(2014) == income_tax_rate_2014, "{} != {}".format(
        parameters.taxes.income_tax_rate(2014), income_tax_rate_2014)
    assert parameters.taxes.income_tax_rate(2015) == income_tax_rate_2014, "{} != {}".format(
        parameters.taxes.income_tax_rate(2015), income_tax_rate_2014)


def test_asof_scale_parameters():
    """
        Test parameters_asof on a scale parameter
    """
    tax_benefit_system = CountryTaxBenefitSystem()
    parameters = tax_benefit_system.parameters
    social_security_contribution_2016 = parameters.taxes.social_security_contribution(2016).thresholds[1]
    social_security_contribution_2017 = parameters.taxes.social_security_contribution(2017).thresholds[1]
    assert social_security_contribution_2016 != social_security_contribution_2017
    parameters_asof(parameters, instant = "2016")
    assert parameters.taxes.social_security_contribution(2016).thresholds[1] == social_security_contribution_2016
    assert parameters.taxes.social_security_contribution(2017).thresholds[1] == social_security_contribution_2016


def test_inflate_simple_parameter():
    """
        Test parameters inflator on a simple parameter as the basic income
    """
    tax_benefit_system = CountryTaxBenefitSystem()
    parameters = tax_benefit_system.parameters
    basic_income_2016 = parameters.benefits.basic_income(2016)
    basic_income_2017 = parameters.benefits.basic_income(2017)
    assert basic_income_2017 == basic_income_2016
    inflate_parameters(parameters, inflator = .1, base_year = 2016, last_year = 2017)

    assert basic_income_2016 == parameters.benefits.basic_income(2016)
    assert 1.1 * basic_income_2016 == parameters.benefits.basic_income(2017)


def test_inflate_scale():
    """
        Test parameters inflator on a scale parameter as the social security contributions tax_scale
    """
    tax_benefit_system = CountryTaxBenefitSystem()
    parameters = tax_benefit_system.parameters
    inflate_parameters(parameters, inflator = .3, base_year = 2015, last_year = 2016)
    for (threshold_2016, threshold_2015) in zip(
            parameters.taxes.social_security_contribution(2016).thresholds,
            parameters.taxes.social_security_contribution(2015).thresholds
            ):
        assert threshold_2016 == threshold_2015 * 1.3


def test_inflate_scale_with_changing_number_of_brackets():
    """
        Test parameters inflator on a scale parameter when the number of brackets changes

        Use parameters_asof to use the present legislation the future pre-inflated legislation
        Test on the social security contributions tax_scale
    """
    tax_benefit_system = CountryTaxBenefitSystem()
    parameters = tax_benefit_system.parameters
    parameters_asof(parameters, instant = periods.instant(2016))  # Remove post 2016 legislation changes
    inflate_parameters(parameters, inflator = .3, base_year = 2016, last_year = 2017)
    for (threshold_2017, threshold_2016) in zip(
            parameters.taxes.social_security_contribution(2017).thresholds,
            parameters.taxes.social_security_contribution(2016).thresholds
            ):
        assert threshold_2017 == threshold_2016 * 1.3, "{} != {}".format(
            threshold_2017, threshold_2016 * 1.3
            )


if __name__ == '__main__':
    test_inflate_simple_parameter()
    test_inflate_scale()
    test_inflate_scale_with_changing_number_of_brackets()
    test_asof_simple_annual_parameter()
    test_asof_scale_parameters()





import logging
import os
import pkg_resources

import pandas as pd


log = logging.getLogger(__name__)

config_files_directory = os.path.join(
    pkg_resources.get_distribution('openfisca-survey-manager').location)


def nnd_hotdeck_using_feather(receiver = None, donor = None, matching_variables = None, z_variables = None):
    """
    Not working
    """
    import feather

    assert receiver is not None and donor is not None
    assert matching_variables is not None

    temporary_directory_path = os.path.join(config_files_directory, 'tmp')
    assert os.path.exists(temporary_directory_path)
    receiver_path = os.path.join(temporary_directory_path, 'receiver.feather')
    donor_path = os.path.join(temporary_directory_path, 'donor.feather')
    feather.write_dataframe(receiver, receiver_path)
    feather.write_dataframe(donor, donor_path)
    if isinstance(matching_variables, str):
        match_vars = '"{}"'.format(matching_variables)
    elif len(matching_variables) == 1:
        match_vars = '"{}"'.format(matching_variables[0])
    else:
        match_vars = '"{}"'.format('todo')

    r_script = """
rm(list=ls())
gc()
devtools::install_github("wesm/feather/R")
library(feather)
library(StatMatch)

receiver <- read_feather({receiver_path})
donor <- read_feather({donor_path})
summary(receiver)
summary(donor)

# variables
receiver = as.data.frame(receiver)
donor = as.data.frame(donor)
gc()
match_vars = {match_vars}
# don_class = c("sexe")
out.nnd <- NND.hotdeck(
  data.rec = receiver, data.don = donor, match.vars = match_vars
  )

# out.nndsummary(out.nnd$mtc.ids)
# head(out.nnd$mtc.ids, 10)
# head(receiver, 10)

fused.nnd.m <- create.fused(
    data.rec = receiver, data.don = donor,
    mtc.ids = out.nnd$mtc.ids,
    z.vars = "{z_variables}"
    )
summary(fused.nnd.m)
""".format(
        receiver_path = receiver_path,
        donor_path = donor_path,
        match_vars = match_vars,
        z_variables = z_variables,
        )
    print(r_script)


def nnd_hotdeck_using_rpy2(receiver = None, donor = None, matching_variables = None,
        z_variables = None, donor_classes = None):
    from rpy2.robjects.packages import importr
    from rpy2.robjects import pandas2ri

    assert receiver is not None and donor is not None
    assert matching_variables is not None

    pandas2ri.activate()
    StatMatch = importr("StatMatch")

    if isinstance(donor_classes, str):
        assert donor_classes in receiver, 'Donor class not present in receiver'
        assert donor_classes in donor, 'Donor class not present in donor'

    try:
        if donor_classes:
            out_NND = StatMatch.NND_hotdeck(
                data_rec = receiver,
                data_don = donor,
                match_vars = pd.Series(matching_variables),
                don_class = pd.Series(donor_classes)
                )
        else:
            out_NND = StatMatch.NND_hotdeck(
                data_rec = receiver,
                data_don = donor,
                match_vars = pd.Series(matching_variables),
                # don_class = pd.Series(donor_classes)
                )
    except Exception as e:
        print(1)
        print(receiver)
        print(2)
        print(donor)
        print(3)
        print(pd.Series(matching_variables))
        print(e)

    # create synthetic data.set, without the
    # duplication of the matching variables

    fused_0 = pandas2ri.ri2py(
        StatMatch.create_fused(
            data_rec = receiver,
            data_don = donor,
            mtc_ids = out_NND[0],
            z_vars = pd.Series(z_variables)
            )
        )

    # create synthetic data.set, with the "duplication"
    # of the matching variables

    fused_1 = pandas2ri.ri2py(
        StatMatch.create_fused(
            data_rec = receiver,
            data_don = donor,
            mtc_ids = out_NND[0],
            z_vars = pd.Series(z_variables),
            dup_x = True,
            match_vars = pd.Series(matching_variables)
            )
        )

    return fused_0, fused_1


if __name__ == "__main__":
    log.setLevel(logging.INFO)

    receiver = pd.DataFrame()
    donor = pd.DataFrame()
    matching_variables = "sexe"
    z_variables = "ident"

    nnd_hotdeck_using_feather(
        receiver = receiver,
        donor = donor,
        matching_variables = matching_variables,
        z_variables = z_variables,
        )





try:
    import pysal as ps
except ModuleNotFoundError:
    pass
import pandas as pd


def read_dbf(dbf_path, index = None, cols = False, incl_index = False):
    """
    Read a dbf file as a pandas.DataFrame, optionally selecting the index
    variable and which columns are to be loaded.

    __author__  = "Dani Arribas-Bel <darribas@asu.edu> "
    ...

    Arguments
    ---------
    dbf_path    : str
                  Path to the DBF file to be read
    index       : str
                  Name of the column to be used as the index of the DataFrame
    cols        : list
                  List with the names of the columns to be read into the
                  DataFrame. Defaults to False, which reads the whole dbf
    incl_index  : Boolean
                  If True index is included in the DataFrame as a
                  column too. Defaults to False

    Returns
    -------
    df          : DataFrame
                  pandas.DataFrame object created
    """
    db = ps.open(dbf_path)
    if cols:
        if incl_index:
            cols.append(index)
        vars_to_read = cols
    else:
        vars_to_read = db.header
    data = dict([(var, db.by_col(var)) for var in vars_to_read])
    if index:
        index = db.by_col(index)
        db.close()
        return pd.DataFrame(data, index=index)
    else:
        db.close()
        return pd.DataFrame(data)




import codecs
import collections
import json
import os

import logging


from openfisca_survey_manager import default_config_files_directory
from openfisca_survey_manager.surveys import Survey
from openfisca_survey_manager.config import Config


log = logging.getLogger(__name__)


class SurveyCollection(object):
    """A collection of Surveys"""
    config = None
    json_file_path = None
    label = None
    name = None
    surveys = list()

    def __init__(self, config_files_directory = default_config_files_directory, label = None, name = None, json_file_path = None):

        log.debug("Initializing SurveyCollection from config file found in {} ..".format(config_files_directory))
        config = Config(config_files_directory = config_files_directory)
        if label is not None:
            self.label = label
        if name is not None:
            self.name = name
        if json_file_path is not None:
            self.json_file_path = json_file_path
            if 'collections' not in config.sections():
                config["collections"] = dict()
            config.set("collections", self.name, self.json_file_path)
            config.save()
        elif config is not None:
            if config.has_option("collections", self.name):
                self.json_file_path = config.get("collections", self.name)
            elif config.get("collections", 'collections_directory') is not None:
                self.json_file_path = os.path.join(
                    config.get("collections", 'collections_directory'),
                    name + '.json',
                    )

        self.config = config

    def __repr__(self):
        header = """{}
Survey collection of {}
Contains the following surveys :
""".format(self.name, self.label)
        surveys = ["       {} : {} \n".format(survey.name, survey.label) for survey in self.surveys]
        return header + "".join(surveys)

    def dump(self, config_files_directory = None, json_file_path = None):
        if self.config is not None:
            config = self.config
        else:
            if config_files_directory is not None:
                pass
            else:
                config_files_directory = default_config_files_directory
            self.config = Config(config_files_directory = config_files_directory)

        if json_file_path is None:
            assert self.json_file_path is not None, 'A json_file_path shoud be provided'
        else:
            self.json_file_path = json_file_path

        config.set("collections", self.name, self.json_file_path)
        config.save
        config.save()
        with codecs.open(self.json_file_path, 'w', encoding = 'utf-8') as _file:
            json.dump(self.to_json(), _file, ensure_ascii = False, indent = 2)

    def fill_hdf(self, source_format = None, surveys = None, tables = None, overwrite = False):
        if source_format is not None:
            assert source_format in ["csv", "Rdata", "sas", "spss", "stata"], \
                "Data source format {} is unknown".format(source_format)
        if surveys is None:
            surveys = self.surveys
        for survey in surveys:
            survey.fill_hdf(source_format = source_format, tables = tables, overwrite = overwrite)
        self.dump()

    def get_survey(self, survey_name):
        available_surveys_names = [survey.name for survey in self.surveys]
        assert survey_name in available_surveys_names, \
            'Survey {} cannot be found for survey collection {}.\nAvailable surveys are :{}'.format(
                survey_name, self.name, available_surveys_names)
        return [survey for survey in self.surveys if survey.name == survey_name].pop()

    @classmethod
    def load(cls, json_file_path = None, collection = None, config_files_directory = default_config_files_directory):
        assert os.path.exists(config_files_directory)
        if json_file_path is None:
            assert collection is not None, "A collection is needed"
            config = Config(config_files_directory = config_files_directory)
            try:
                json_file_path = config.get("collections", collection)
            except Exception as error:
                log.debug("Looking for congi file in {}".format(config_files_directory))
                log.error(error)
                raise

        with open(json_file_path, 'r') as _file:
            self_json = json.load(_file)
            name = self_json.get('name')

        self = cls(name = name)
        self.config = config
        with open(json_file_path, 'r') as _file:
            self_json = json.load(_file)
            self.json_file_path = json_file_path
            self.label = self_json.get('label')
            self.name = self_json.get('name')

        surveys = self_json.get('surveys')
        for survey_name, survey_json in surveys.items():
            survey = Survey(name = survey_name)
            self.surveys.append(survey.create_from_json(survey_json))
        return self

    def to_json(self):
        self_json = collections.OrderedDict((
            ))
        self_json['name'] = self.name
        self_json['surveys'] = collections.OrderedDict((
            ))
        for survey in self.surveys:
            self_json['surveys'][survey.name] = survey.to_json()
        return self_json



import logging


from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario

log = logging.getLogger(__name__)


def test_create_data_frame_by_entity():
    survey_scenario = create_randomly_initialized_survey_scenario()
    period = '2017-01'
    df_by_entity = survey_scenario.create_data_frame_by_entity(variables = ['salary', 'rent'], period = period)
    salary = survey_scenario.calculate_variable('salary', period = period)
    rent = survey_scenario.calculate_variable('rent', period = period)
    for entity, df in df_by_entity.items():
        assert not df.empty, "{} dataframe is empty".format(entity)
    assert (df_by_entity['person']['salary'] == salary).all().all()
    assert (df_by_entity['household']['rent'] == rent).all().all()



import numpy as np

from openfisca_survey_manager.statshelpers import bottom_share, top_share


size = 1000
x = np.ones(size) + np.random.uniform(0, .00000001, size)


def test_bottom_share():
    np.testing.assert_almost_equal(
        bottom_share(x, .4),
        .4,
        )


def test_to_share():
    np.testing.assert_almost_equal(
        top_share(x, .1),
        .1
        )





import logging
import os
import pkg_resources


log = logging.getLogger(__name__)

# Hack to for use at the CASD (shared user)
# Use taxipp/.config/ directory if exists as default_config_files_directory
try:
    taxipp_location = pkg_resources.get_distribution('taxipp').location
    default_config_files_directory = os.path.join(taxipp_location, '.config', 'openfisca-survey-manager')
except pkg_resources.DistributionNotFound:
    taxipp_location = None

if taxipp_location is None or not os.path.exists(default_config_files_directory):
    default_config_files_directory = None

# Travis tests
is_travis = 'TRAVIS' in os.environ
is_circleci = 'CIRCLECI' in os.environ

test_config_files_directory = os.path.join(
    pkg_resources.get_distribution('openfisca-survey-manager').location,
    'openfisca_survey_manager',
    'tests',
    'data_files',
    )
with open(os.path.join(test_config_files_directory, 'config_template.ini')) as file:
    config_ini = file.read()

config_ini = config_ini.format(location = pkg_resources.get_distribution('openfisca-survey-manager').location)
with open(os.path.join(test_config_files_directory, 'config.ini'), "w+") as file:
    file.write(config_ini)

if is_travis or is_circleci:
    default_config_files_directory = test_config_files_directory

if default_config_files_directory is None:
    from xdg import BaseDirectory
    default_config_files_directory = BaseDirectory.save_config_path('openfisca-survey-manager')

    log.debug('Using default_config_files_directory = {}'.format(
        default_config_files_directory
        ))





import logging

from openfisca_country_template import CountryTaxBenefitSystem
from openfisca_survey_manager.tests.test_scenario import create_randomly_initialized_survey_scenario

log = logging.getLogger(__name__)


tax_benefit_system = CountryTaxBenefitSystem()


def test_summarize_variable():
    survey_scenario = create_randomly_initialized_survey_scenario()
    survey_scenario.summarize_variable(variable = "rent", force_compute = True)
    survey_scenario.summarize_variable(variable = "housing_occupancy_status", force_compute = True)


if __name__ == "__main__":
    # log = logging.getLogger(__name__)
    # logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)
    test_summarize_variable()





import logging
import operator

from numpy import exp, ones, zeros, unique, array, dot, float64


log = logging.getLogger(__name__)


def linear(u):
    """

    Args:
      u:

    Returns:

    """
    return 1 + u


def linear_prime(u):
    """

    Args:
      u:

    Returns:

    """
    return ones(u.shape, dtype = float)


def raking_ratio(u):
    """

    Args:
      u:

    Returns:

    """
    return exp(u)


def raking_ratio_prime(u):
    """

    Args:
      u:

    Returns:

    """
    return exp(u)


def logit(u, low, up):
    """

    Args:
      u:
      low:
      up:

    Returns:

    """
    a = (up - low) / ((1 - low) * (up - 1))
    return (low * (up - 1) + up * (1 - low) * exp(a * u)) / (up - 1 + (1 - low) * exp(a * u))


def logit_prime(u, low, up):
    """

    Args:
      u:
      low:
      up:

    Returns:

    """
    a = (up - low) / ((1 - low) * (up - 1))
    return (
        (a * up * (1 - low) * exp(a * u)) * (up - 1 + (1 - low) * exp(a * u))
        - (low * (up - 1) + up * (1 - low) * exp(a * u)) * (1 - low) * a * exp(a * u)
        ) / (up - 1 + (1 - low) * exp(a * u)) ** 2


def build_dummies_dict(data):
    """

    Args:
      data:

    Returns:


    """
    unique_val_list = unique(data)
    output = {}
    for val in unique_val_list:
        output[val] = (data == val)
    return output


def calmar(data_in, margins, initial_weight = 'wprm_init', method = 'linear', lo = None, up = None, use_proportions = False,
        xtol = 1.49012e-08, maxfev = 256):
    """Calibrates weights to satisfy margins constraints

    Args:
        data_in (pd.DataFrame): The observations data
        margins (dict): Margins is a dictionnary containing for each variable as key the following values
          - a scalar for numeric variables
          - a dictionnary with categories as key and populations as values
          - eventually a key named `total_population` with value the total population. If absent it is initialized to the actual total population
        initial_weight (str, optional): Initial weight variable. Defaults to 'wprm_init'.
        method (str, optional): Calibration method. Should be 'linear', 'raking ratio' or 'logit'. Defaults to 'linear'.
        lo (float, optional): Lower bound on weights ratio. Mandatory when using logit method. Should be < 1. Defaults to None.
        up (float, optional): Upper bound on weights ratio. Mandatory when using logit method. Should be > 1. Defaults to None.
        use_proportions (bool, optional): When True use proportions if total population from margins doesn't match total population. Defaults to False.
        xtol (float, optional): Relative precision on lagrangian multipliers.  Defaults to 1.49012e-08 (fsolve xtol).
        maxfev (int, optional): Maximum number of function evaluation. Defaults to 256.

    Raises:
        Exception: [description]
        Exception: [description]
        Exception: [description]

    Returns:
        np.array: Margins adjusting weights
        float: Lagrangian parameter
        dict: Updated margins
    """
    from scipy.optimize import fsolve

    # remove null weights and keep original data
    null_weight_observations = data_in[initial_weight].isnull().sum()
    if null_weight_observations > 0:
        log.info("{} observations have a NaN weight. Not used in the calibration.".format(null_weight_observations))

    is_non_zero_weight = (data_in[initial_weight].fillna(0) > 0)
    if is_non_zero_weight.sum() > null_weight_observations:
        log.info("{} observations have a zero weight. Not used in the calibration.".format(
            (data_in[initial_weight].fillna(0) <= 0).sum() - null_weight_observations))

    variables = set(margins.keys()).intersection(set(data_in.columns))
    for variable in variables:
        null_value_observations = data_in[variable].isnull().sum()
        if null_value_observations > 0:
            log.info("For variable {}, {} observations have a NaN value. Not used in the calibration.".format(
                variable, null_value_observations))
            is_non_zero_weight = is_non_zero_weight & data_in[variable].notnull()

    if not is_non_zero_weight.all():
        log.info("We drop {} observations.".format((~is_non_zero_weight).sum()))

    data = dict()
    for a in data_in.columns:
        data[a] = data_in.loc[is_non_zero_weight, a].copy()

    if not margins:
        raise Exception("Calmar requires non empty dict of margins")

    # choose method
    assert method in ['linear', 'raking ratio', 'logit'], "method should be 'linear', 'raking ratio' or 'logit'"
    if method == 'linear':
        F = linear
        F_prime = linear_prime
    elif method == 'raking ratio':
        F = raking_ratio
        F_prime = raking_ratio_prime
    elif method == 'logit':
        assert up is not None, "When method == 'logit', a value > 1 for up is mandatory"
        assert up > 1, "up should be > 1"
        assert lo is not None, "When method == 'logit', a value < 1 for lo is mandatory"
        assert lo < 1, "lo should be < 1"

        def F(x):
            return logit(x, lo, up)

        def F_prime(x):
            return logit_prime(x, lo, up)

    # Construction observations matrix
    if 'total_population' in margins:
        total_population = margins.pop('total_population')
    else:
        total_population = data[initial_weight].fillna(0).sum()

    nk = len(data[initial_weight])
    # number of Lagrange parameters (at least total population)
    nj = 1

    margins_new = {}
    margins_new_dict = {}
    for var, val in margins.items():
        if isinstance(val, dict):
            dummies_dict = build_dummies_dict(data[var])
            k, pop = 0, 0
            for cat, nb in val.items():
                cat_varname = var + '_' + str(cat)
                data[cat_varname] = dummies_dict[cat]
                margins_new[cat_varname] = nb
                if var not in margins_new_dict:
                    margins_new_dict[var] = {}
                margins_new_dict[var][cat] = nb
                pop += nb
                k += 1
                nj += 1
            # Check total popualtion
            if pop != total_population:
                if use_proportions:
                    log.info(
                        'calmar: categorical variable {} is inconsistent with population; using proportions'.format(
                            var
                            )
                        )
                    for cat, nb in val.items():
                        cat_varname = var + '_' + str(cat)
                        margins_new[cat_varname] = nb * total_population / pop
                        margins_new_dict[var][cat] = nb * total_population / pop
                else:
                    raise Exception('calmar: categorical variable {} weights sums up to {} != {}'.format(
                        var, pop, total_population))
        else:
            margins_new[var] = val
            margins_new_dict[var] = val
            nj += 1

    # On conserve systematiquement la population
    if hasattr(data, 'dummy_is_in_pop'):
        raise Exception('dummy_is_in_pop is not a valid variable name')

    data['dummy_is_in_pop'] = ones(nk)
    margins_new['dummy_is_in_pop'] = total_population

    # paramètres de Lagrange initialisés à zéro
    lambda0 = zeros(nj)

    # initial weights
    d = data[initial_weight].values
    x = zeros((nk, nj))  # nb obs x nb constraints
    xmargins = zeros(nj)
    margins_dict = {}
    j = 0
    for var, val in margins_new.items():
        x[:, j] = data[var]
        xmargins[j] = val
        margins_dict[var] = val
        j += 1

    # Résolution des équations du premier ordre
    def constraint(lambda_):
        return dot(d * F(dot(x, lambda_)), x) - xmargins

    def constraint_prime(lambda_):
        return dot(d * (x.T * F_prime(dot(x, lambda_))), x)
        # le jacobien ci-dessus est constraintprime = @(lambda) x*(d.*Fprime(x'*lambda)*x');

    tries, ier = 0, 2
    err_max = 1
    conv = 1
    while (ier == 2 or ier == 5 or ier == 4) and not (tries >= 10 or (err_max < 1e-6 and conv < 1e-8)):
        lambdasol, infodict, ier, mesg = fsolve(
            constraint,
            lambda0,
            fprime = constraint_prime,
            maxfev = maxfev,
            xtol = xtol,
            full_output = 1,
            )
        lambda0 = 1 * lambdasol
        tries += 1

        pondfin = d * F(dot(x, lambdasol))
        rel_error = {}
        for var, val in margins_new.items():  # noqa analysis:ignore
            rel_error[var] = abs((data[var] * pondfin).sum() - margins_dict[var]) / margins_dict[var]
        sorted_err = sorted(rel_error.items(), key = operator.itemgetter(1), reverse = True)

        conv = abs(err_max - sorted_err[0][1])
        err_max = sorted_err[0][1]

    if (ier == 2 or ier == 5 or ier == 4):
        log.debug("optimization converged after {} tries".format(tries))

    # rebuilding a weight vector with the same size of the initial one
    pondfin_out = array(data_in[initial_weight], dtype = float64)
    pondfin_out[is_non_zero_weight] = pondfin

    del infodict, mesg  # TODO better exploit this information

    return pondfin_out, lambdasol, margins_new_dict


def check_calmar(data_in, margins, initial_weight='wprm_init', pondfin_out = None, lambdasol = None, margins_new_dict = None):
    """

    Args:
      data_in:
      margins:
      initial_weight:  (Default value = 'wprm_init')
      pondfin_out:  (Default value = None)
      lambdasol:  (Default value = None)
      margins_new_dict:  (Default value = None)

    Returns:

    """
    for variable, margin in margins.items():
        if variable != 'total_population':
            print(variable, margin, abs(margin - margins_new_dict[variable]) / abs(margin))





"""
Example from page 23 of calmar documentation
https://www.insee.fr/fr/statistiques/fichier/2021902/doccalmar.pdf
"""

import numpy as np
import pandas as pd


from openfisca_survey_manager.calmar import calmar


def create_input_dataframe():
    columns = ['X', 'Y', 'Z', 'POND']
    index = [
        'A',
        'B',
        'C',
        'D',
        'E',
        'F',
        'G',
        'H',
        'I',
        'J',
        'K',
        ]
    df = pd.DataFrame(columns = columns, index = index)
    values_by_index = {
        'A': [1, 1, 1, 10],
        'B': [1, 2, 2, 0],
        'C': [1, 2, 3, np.nan],
        'D': [2, 1, 1, 11],
        'E': [2, 1, 3, 13],
        'F': [2, 2, 2, 7],
        'G': [2, 2, 2, 8],
        'H': [1, 2, 2, 8],
        'I': [2, 1, 2, 9],
        'J': [np.nan, 2, 2, 10],
        'K': [2, 2, 2, 14],
        }
    for index, values in values_by_index.items():
        df.loc[index] = pd.Series(dict(zip(columns, values)))

    df['Z'] = df.Z.astype(float)
    df['POND'] = df.POND.astype(float)
    return df


def create_margins():
    margins_by_variable = {
        'X': {
            1: 20,
            2: 60,
            },
        'Y': {
            1: 30,
            2: 50,
            },
        'Z': 140.0
        }
    return margins_by_variable


def test_calmar():
    target_weight_ratio = pd.Series([
        1.01683,
        np.nan,
        1.22897,
        np.nan,
        1.14602,
        0.49456,
        0.21342,
        1.38511,
        1.38511,
        1.38511,
        1.00000,
        ])

    data = create_input_dataframe()
    margins_by_variable = create_margins()
    pondfin_out, lambdasol, margins_new_dict = calmar(data, margins_by_variable, method = 'raking ratio', initial_weight = 'POND')

    data['weightt_ratio'] = pondfin_out / data.POND
    weight_ratio = data.sort_values(['X', 'Y', 'Z'])['weightt_ratio'].round(5)
    null_target_weight_ratio = target_weight_ratio.isnull()

    assert weight_ratio.loc[null_target_weight_ratio.values].isnull().all(), "Error on Nan"

    assert (
        target_weight_ratio.loc[~null_target_weight_ratio.values].values
        == weight_ratio.loc[~null_target_weight_ratio.values].values
        ).all(), "Erros on non NaN values"


if __name__ == '__main__':
    import logging
    import sys
    log = logging.getLogger(__name__)
    verbose = True
    logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)
    test_calmar()



import collections
from datetime import datetime
import logging
import os

import numpy as np
import pandas as pd


log = logging.getLogger(__name__)


# TODO:
#  * Localisation

class AbstractAggregates(object):
    aggregate_variables = None
    amount_unit = 1e6
    currency = None
    base_data_frame = None
    baseline_simulation = None
    beneficiaries_unit = 1e3
    filter_by = None
    labels = None
    simulation = None
    survey_scenario = None
    totals_df = None

    def __init__(self, survey_scenario = None):
        assert survey_scenario is not None
        self.year = survey_scenario.year
        self.survey_scenario = survey_scenario
        assert self.simulation is None

        assert survey_scenario.simulation is not None
        self.simulation = survey_scenario.simulation

        if survey_scenario.baseline_tax_benefit_system is not None:
            assert survey_scenario.baseline_simulation is not None
            self.baseline_simulation = survey_scenario.baseline_simulation
        else:
            self.baseline_simulation = None

        self.weight_variable_by_entity = survey_scenario.weight_variable_by_entity
        if self.labels is None:
            amount_unit_str = "({} {})".format(self.amount_unit, self.currency)
            beneficiaries_unit_str = "({})".format(self.beneficiaries_unit)
            self.labels = collections.OrderedDict((
                ('label', "Mesure"),
                ('entity', "Entité"),
                ('reform_amount', "Dépenses\n" + amount_unit_str),
                ('reform_beneficiaries', "Bénéficiaires\n(milliers)"),
                ('baseline_amount', "Dépenses initiales\n" + amount_unit_str),
                ('baseline_beneficiaries', "Bénéficiaires\ninitiaux\n" + beneficiaries_unit_str),
                ('actual_amount', "Dépenses\nréelles\n" + amount_unit_str),
                ('actual_beneficiaries', "Bénéficiaires\nréels\n" + beneficiaries_unit_str),
                ('amount_absolute_difference', "Diff. absolue\nDépenses\n" + amount_unit_str),
                ('beneficiaries_absolute_difference', "Diff absolue\nBénéficiaires\n" + beneficiaries_unit_str),
                ('amount_relative_difference', "Diff. relative\nDépenses"),
                ('beneficiaries_relative_difference', "Diff. relative\nBénéficiaires"),
                ))

    def compute_aggregates(self, use_baseline = True, reform = True, actual = True):
        """
        Compute aggregate amounts
        """
        filter_by = self.filter_by
        self.totals_df = self.load_actual_data(year = self.year)

        simulation_types = list()
        if use_baseline:
            assert self.baseline_simulation is not None
            simulation_types.append('baseline')
        if reform:
            simulation_types.append('reform')
        if actual:
            simulation_types.append('actual')

        data_frame_by_simulation_type = dict()

        for simulation_type in simulation_types:
            if simulation_type == 'actual':
                data_frame_by_simulation_type['actual'] = self.totals_df.copy() if self.totals_df is not None else None
            else:
                use_baseline = False if simulation_type == 'reform' else True
                data_frame = pd.DataFrame()
                for variable in self.aggregate_variables:
                    variable_data_frame = self.compute_variable_aggregates(
                        variable, use_baseline = use_baseline, filter_by = filter_by)
                    data_frame = pd.concat((data_frame, variable_data_frame))

                data_frame.rename(columns = {
                    'amount': '{}_amount'.format(simulation_type),
                    'beneficiaries': '{}_beneficiaries'.format(simulation_type),
                    },
                    inplace = True
                    )
                data_frame_by_simulation_type[simulation_type] = data_frame

        if use_baseline and reform:
            del data_frame_by_simulation_type['reform']['entity']
            del data_frame_by_simulation_type['reform']['label']

        self.base_data_frame = pd.concat(
            list(data_frame_by_simulation_type.values()),
            axis = 1,
            sort = True,
            ).loc[self.aggregate_variables]
        return self.base_data_frame

    def compute_difference(self, target = "baseline", default = 'actual', amount = True, beneficiaries = True,
            absolute = True, relative = True):
        """Computes and add relative and/or absolute differences to the data_frame."""
        assert relative or absolute
        assert amount or beneficiaries
        base_data_frame = self.base_data_frame if self.base_data_frame is not None else self.compute_aggregates()

        difference_data_frame = base_data_frame[['label', 'entity']].copy()
        quantities = list()
        quantities += ['amount'] if amount else None
        quantities += ['beneficiaries'] if beneficiaries else None

        try:
            for quantity in quantities:
                difference_data_frame['{}_absolute_difference'.format(quantity)] = (
                    abs(base_data_frame['{}_{}'.format(target, quantity)]) - base_data_frame['{}_{}'.format(default, quantity)]
                    )
                difference_data_frame['{}_relative_difference'.format(quantity)] = (
                    abs(base_data_frame['{}_{}'.format(target, quantity)]) - base_data_frame['{}_{}'.format(default, quantity)]
                    ) / abs(base_data_frame['{}_{}'.format(default, quantity)])
        except KeyError as e:
            log.debug(e)
            log.debug("Do not computing differences")
            return None

        return difference_data_frame

    def compute_variable_aggregates(self, variable, use_baseline = False, filter_by = None):
        """Returns aggregate spending, and number of beneficiaries for the relevant entity level.

        Parameters
        ----------
        variable : string
                   name of the variable aggregated according to its entity
        use_baseline : bool
                    Use the baseline or the reform or the only avalilable simulation when no reform (default)
        filter_by : string or boolean
                    If string use it as the name of the variable to filter by
                    If not None or False and the string is not present in the tax-benefit-system use the default filtering variable if any
        """
        if use_baseline:
            simulation = self.baseline_simulation
        else:
            simulation = self.simulation

        variables = simulation.tax_benefit_system.variables
        column = variables.get(variable)

        if column is None:
            msg = "Variable {} is not available".format(variable)
            if use_baseline:
                msg += " in baseline simulation"
            log.info(msg)
            return pd.DataFrame(
                data = {
                    'label': variable,
                    'entity': 'Unknown entity',
                    'amount': 0,
                    'beneficiaries': 0,
                    },
                index = [variable],
                )
        weight = self.weight_variable_by_entity[column.entity.key]
        assert weight in variables, "{} not a variable of the tax_benefit_system".format(weight)

        weight_array = simulation.calculate(weight, period = self.year).astype('float')
        assert not np.isnan(np.sum(weight_array)), "The are some NaN in weights {} for entity {}".format(
            weight, column.entity.key)
        # amounts and beneficiaries from current data and default data if exists
        # Build weights for each entity
        variable_array = simulation.calculate_add(variable, period = self.year).astype('float')
        assert np.isfinite(variable_array).all(), "The are non finite values in variable {} for entity {}".format(
            variable, column.entity.key)
        data = pd.DataFrame({
            variable: variable_array,
            weight: weight_array,
            })
        if filter_by:
            filter_dummy_variable = (
                filter_by
                if filter_by in variables
                else self.survey_scenario.filtering_variable_by_entity[column.entity.key]
                )
            filter_dummy_array = simulation.calculate(filter_dummy_variable, period = self.year)

        else:
            filter_dummy_array = 1

        assert np.isfinite(filter_dummy_array).all(), "The are non finite values in variable {} for entity {}".format(
            filter_dummy_variable, column.entity.key)

        amount = int(
            (
                data[variable]
                * data[weight]
                * filter_dummy_array
                / self.amount_unit
                ).sum()
            )
        beneficiaries = int(
            (
                (data[variable] != 0)
                * data[weight]
                * filter_dummy_array
                / self.beneficiaries_unit
                ).sum()
            )
        variable_data_frame = pd.DataFrame(
            data = {
                'label': variables[variable].label,
                'entity': variables[variable].entity.key,
                'amount': amount,
                'beneficiaries': beneficiaries,
                },
            index = [variable],
            )

        return variable_data_frame

    def create_description(self):
        """
        Create a description dataframe
        """
        now = datetime.now()
        return pd.DataFrame([
            'OpenFisca',
            'Calculé le %s à %s' % (now.strftime('%d-%m-%Y'), now.strftime('%H:%M')),
            'Système socio-fiscal au %s' % self.simulation.period.start.year,
            "Données d'enquêtes de l'année %s" % str(self.data_year),
            ])

    def to_csv(self, path = None, absolute = True, amount = True, beneficiaries = True, default = 'actual',
            relative = True, target = "reform"):
        """Saves the table to csv."""
        assert path is not None

        if os.path.isdir(path):
            now = datetime.now()
            file_path = os.path.join(path, 'Aggregates_%s.%s' % (now.strftime('%d-%m-%Y'), ".csv"))
        else:
            file_path = path

        df = self.get_data_frame(
            absolute = absolute,
            amount = amount,
            beneficiaries = beneficiaries,
            default = default,
            relative = relative,
            target = target,
            )
        df.to_csv(file_path, index = False, header = True)

    def to_excel(self, path = None, absolute = True, amount = True, beneficiaries = True, default = 'actual',
            relative = True, target = "reform"):
        """Saves the table to excel."""
        assert path is not None

        if os.path.isdir(path):
            now = datetime.now()
            file_path = os.path.join(path, 'Aggregates_%s.%s' % (now.strftime('%d-%m-%Y'), ".xlsx"))
        else:
            file_path = path

        df = self.get_data_frame(
            absolute = absolute,
            amount = amount,
            beneficiaries = beneficiaries,
            default = default,
            relative = relative,
            target = target,
            )
        writer = pd.ExcelWriter(file_path)
        df.to_excel(writer, "aggregates", index = False, header = True)
        descr = self.create_description()
        descr.to_excel(writer, "description", index = False, header = False)
        writer.save()

    def to_html(self, path = None, absolute = True, amount = True, beneficiaries = True, default = 'actual',
            relative = True, target = "reform"):
        """Gets or saves the table to html format."""
        df = self.get_data_frame(
            absolute = absolute,
            amount = amount,
            beneficiaries = beneficiaries,
            default = default,
            relative = relative,
            target = target,
            )

        if path is not None and os.path.isdir(path):
            now = datetime.now()
            file_path = os.path.join(path, 'Aggregates_%s.%s' % (now.strftime('%d-%m-%Y'), ".html"))
        else:
            file_path = path

        if file_path is not None:
            with open(file_path, "w") as html_file:
                df.to_html(html_file)
        return df.to_html()

    def to_markdown(self, path = None, absolute = True, amount = True, beneficiaries = True, default = 'actual',
            relative = True, target = "reform"):
        """Gets or saves the table to markdown format."""
        df = self.get_data_frame(
            absolute = absolute,
            amount = amount,
            beneficiaries = beneficiaries,
            default = default,
            relative = relative,
            target = target,
            )

        if path is not None and os.path.isdir(path):
            now = datetime.now()
            file_path = os.path.join(path, 'Aggregates_%s.%s' % (now.strftime('%d-%m-%Y'), ".md"))
        else:
            file_path = path

        if file_path is not None:
            with open(file_path, "w") as markdown_file:
                df.to_markdown(markdown_file)

        return df.to_markdown()

    def get_calibration_coeffcient(self, target = "reform"):
        df = self.compute_aggregates(
            actual = True,
            use_baseline = 'baseline' == target,
            reform = 'reform' == target,
            )
        return df['{}_amount'.format(target)] / df['actual_amount']

    def get_data_frame(
            self,
            absolute = True,
            amount = True,
            beneficiaries = True,
            default = 'actual',
            formatting = True,
            relative = True,
            target = "reform",
            ):
        assert target is None or target in ['reform', 'baseline']

        columns = self.labels.keys()
        if (absolute or relative) and (target != default):
            difference_data_frame = self.compute_difference(
                absolute = absolute,
                amount = amount,
                beneficiaries = beneficiaries,
                default = default,
                relative = relative,
                target = target,
                )
        else:
            difference_data_frame = None

        # Removing unwanted columns
        if amount is False:
            columns = [column for column in columns if 'amount' not in columns]

        if beneficiaries is False:
            columns = [column for column in columns if 'beneficiaries' not in column]

        if absolute is False:
            columns = [column for column in columns if 'absolute' not in column]

        if relative is False:
            columns = [column for column in columns if 'relative' not in column]

        for simulation_type in ['reform', 'baseline', 'actual']:
            if simulation_type not in [target, default]:
                columns = [column for column in columns if simulation_type not in column]

        aggregates_data_frame = self.compute_aggregates(
            actual = 'actual' in [target, default],
            use_baseline = 'baseline' in [target, default],
            reform = 'reform' in [target, default],
            )
        ordered_columns = [
            'label',
            'entity',
            'reform_amount',
            'baseline_amount',
            'actual_amount',
            'amount_absolute_difference',
            'amount_relative_difference',
            'reform_beneficiaries',
            'baseline_beneficiaries',
            'actual_beneficiaries',
            'beneficiaries_absolute_difference',
            'beneficiaries_relative_difference'
            ]
        if difference_data_frame is not None:
            df = aggregates_data_frame.merge(difference_data_frame, how = 'left')[columns]
        else:
            columns = [column for column in columns if column in aggregates_data_frame.columns]
            df = aggregates_data_frame[columns]

        df = df.reindex(columns = ordered_columns).dropna(axis = 1, how = 'all').rename(columns = self.labels)

        if formatting:
            relative_columns = [column for column in df.columns if 'relative' in column]
            df[relative_columns] = df[relative_columns].applymap(
                lambda x: "{:.2%}".format(x) if str(x) != 'nan' else 'nan'
                )
            for column in df.columns:
                if issubclass(np.dtype(df[column]).type, np.number):
                    df[column] = (
                        df[column]
                        .apply(lambda x: "{:d}".format(int(round(x))) if str(x) != 'nan' else 'nan')
                        )
        return df

    def load_actual_data(self, year = None):
        NotImplementedError




import logging


from openfisca_core.tools import assert_near
from openfisca_country_template import CountryTaxBenefitSystem
from openfisca_survey_manager.tests.test_scenario import (
    create_randomly_initialized_survey_scenario
    )


log = logging.getLogger(__name__)


tax_benefit_system = CountryTaxBenefitSystem()


def test_compute_marginal_tax_rate():
    survey_scenario = create_randomly_initialized_survey_scenario(use_marginal_tax_rate = True)
    assert survey_scenario._modified_simulation is not None
    assert_near(
        survey_scenario.compute_marginal_tax_rate(target_variable = 'income_tax', period = 2017),
        (1 - .15),
        relative_error_margin = 1e-6,
        )
    survey_scenario.compute_marginal_tax_rate(target_variable = 'disposable_income', period = 2017)


if __name__ == "__main__":
    import sys
    log = logging.getLogger(__name__)
    logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)
    test_compute_marginal_tax_rate()





import logging
import os
import pandas as pd


from openfisca_core import periods
from openfisca_core.parameters import ParameterNode, Scale


log = logging.getLogger(__name__)


def inflate_parameters(parameters, inflator, base_year, last_year = None, ignore_missing_units = False):

    if (last_year is not None) and (last_year > base_year + 1):
        for year in range(base_year + 1, last_year + 1):
            inflate_parameters(parameters, inflator, year - 1, last_year = year, ignore_missing_units = ignore_missing_units)

    else:
        if last_year is None:
            last_year = base_year + 1

        assert last_year == base_year + 1

        for sub_parameter in parameters.children.values():
            if isinstance(sub_parameter, ParameterNode):
                inflate_parameters(sub_parameter, inflator, base_year, last_year, ignore_missing_units = ignore_missing_units)
            else:
                acceptable_units = [
                    'rate_unit',
                    'threshold_unit',
                    'unit',
                    ]
                if ignore_missing_units:
                    if not hasattr(sub_parameter, 'metadata'):
                        continue
                    # Empty intersection: not unit present in metadata
                    if not bool(set(sub_parameter.metadata.keys()) & set(acceptable_units)):
                        continue

                assert hasattr(sub_parameter, 'metadata'), "{} doesn't have metadata".format(sub_parameter.name)
                unit_types = set(sub_parameter.metadata.keys()).intersection(set(acceptable_units))
                assert unit_types, "No admissible unit in metadata for parameter {}. You may consider using the option 'ignore_missing_units' from the inflate_paramaters() function.".format(
                    sub_parameter.name)
                if len(unit_types) > 1:
                    assert unit_types == set(['threshold_unit', 'rate_unit']), \
                        "Too much admissible units in metadata for parameter {}".format(
                            sub_parameter.name)
                unit_by_type = dict([
                    (unit_type, sub_parameter.metadata[unit_type]) for unit_type in unit_types
                    ])

                for unit_type in unit_by_type.keys():
                    if sub_parameter.metadata[unit_type].startswith("currency"):
                        inflate_parameter_leaf(sub_parameter, base_year, inflator, unit_type = unit_type)


def inflate_parameter_leaf(sub_parameter, base_year, inflator, unit_type = 'unit'):
    """
    Inflate a Parameter leaf according to unit type

    Basic unit type are supposed by default
    Other admissible unit types are threshold_unit and rate_unit
    """

    if isinstance(sub_parameter, Scale):
        if unit_type == 'threshold_unit':
            for bracket in sub_parameter.brackets:
                threshold = bracket.children['threshold']
                inflate_parameter_leaf(threshold, base_year, inflator)
            return
    else:
        # Remove new values for year > base_year
        kept_instants_str = [
            parameter_at_instant.instant_str
            for parameter_at_instant in sub_parameter.values_list
            if periods.instant(parameter_at_instant.instant_str).year <= base_year
            ]
        if not kept_instants_str:
            return

        last_admissible_instant_str = max(kept_instants_str)
        sub_parameter.update(
            start = last_admissible_instant_str,
            value = sub_parameter(last_admissible_instant_str)
            )
        restricted_to_base_year_value_list = [
            parameter_at_instant for parameter_at_instant in sub_parameter.values_list
            if periods.instant(parameter_at_instant.instant_str).year == base_year
            ]
        # When value is changed in the base year
        if restricted_to_base_year_value_list:
            for parameter_at_instant in reversed(restricted_to_base_year_value_list):
                if parameter_at_instant.instant_str.startswith(str(base_year)):
                    value = (
                        parameter_at_instant.value * (1 + inflator)
                        if parameter_at_instant.value is not None
                        else None
                        )
                    sub_parameter.update(
                        start = parameter_at_instant.instant_str.replace(
                            str(base_year), str(base_year + 1)
                            ),
                        value = value,
                        )
        # Or use the value at that instant even when it is defined earlier tahn the base year
        else:
            value = (
                sub_parameter("{}-12-31".format(base_year)) * (1 + inflator)
                if sub_parameter("{}-12-31".format(base_year)) is not None
                else None
                )
            sub_parameter.update(
                start = "{}-01-01".format(base_year + 1),
                value = value
                )


def asof(tax_benefit_system, instant):
    parameters = tax_benefit_system.parameters
    parameters_asof(parameters, instant)
    variables_asof(tax_benefit_system, instant)


def leaf_asof(sub_parameter, instant):
    kept_instants_str = [
        parameter_at_instant.instant_str
        for parameter_at_instant in sub_parameter.values_list
        if periods.instant(parameter_at_instant.instant_str) <= instant
        ]
    if not kept_instants_str:
        sub_parameter.values_list = []
        return

    last_admissible_instant_str = max(kept_instants_str)
    sub_parameter.update(
        start = last_admissible_instant_str,
        value = sub_parameter(last_admissible_instant_str)
        )
    return


def parameters_asof(parameters, instant):
    if isinstance(instant, str):
        instant = periods.instant(instant)
    assert isinstance(instant, periods.Instant)

    for sub_parameter in parameters.children.values():
        if isinstance(sub_parameter, ParameterNode):
            parameters_asof(sub_parameter, instant)
        else:
            if isinstance(sub_parameter, Scale):
                for bracket in sub_parameter.brackets:
                    threshold = bracket.children['threshold']
                    rate = bracket.children.get('rate')
                    amount = bracket.children.get('amount')
                    leaf_asof(threshold, instant)
                    if rate:
                        leaf_asof(rate, instant)
                    if amount:
                        leaf_asof(amount, instant)
            else:
                leaf_asof(sub_parameter, instant)


def variables_asof(tax_benefit_system, instant, variables_list = None):
    if isinstance(instant, str):
        instant = periods.instant(instant)
    assert isinstance(instant, periods.Instant)

    if variables_list is None:
        variables_list = tax_benefit_system.variables.keys()

    for variable_name, variable in tax_benefit_system.variables.items():
        if variable_name in variables_list:
            formulas = variable.formulas
            for instant_str in list(formulas.keys()):
                if periods.instant(instant_str) > instant:
                    del formulas[instant_str]

            if variable.end is not None:
                if periods.instant(variable.end) >= instant:
                    variable.end = None


def stata_files_to_data_frames(data, period = None):
    assert period is not None
    period = periods.period(period)

    stata_file_by_entity = data.get('stata_file_by_entity')
    if stata_file_by_entity is None:
        return

    variables_from_stata_files = list()
    input_data_frame_by_entity_by_period = dict()
    input_data_frame_by_entity_by_period[periods.period(period)] = input_data_frame_by_entity = dict()
    for entity, file_path in stata_file_by_entity.items():
        assert os.path.exists(file_path), "Invalid file path: {}".format(file_path)
        entity_data_frame = input_data_frame_by_entity[entity] = pd.read_stata(file_path)
        variables_from_stata_files += list(entity_data_frame.columns)
    data['input_data_frame_by_entity_by_period'] = input_data_frame_by_entity_by_period

    return variables_from_stata_files
